/**********************************************************************
*
*    Author:		shondll (shondll_yahoo.com)
*    Company:		Olimex Ltd.
*    Date:			28/10/2010
*    File Version:	1.00
*    Description:	Initialization routines
*
**********************************************************************/
#include "Console.h"
#include "ConfigApp.h"
#include "TimeDelay.h"
#include "HardwareProfile.h"
#include "WirelessProtocols\MCHP_API.h"
#include "WirelessProtocols\LCDBlocking.h"

#include "WirelessProtocols\SymbolTime.h"
#include "TCN75A.h"

// this line enbales displaying of data on the 1x9 LCD of MOD-LCD1x9 if one is connected to a rev.B MOD-ZIGBEE board
// you want get any runtime errors if you leave this enabled because data is only sent to the LCD and not acknowledged,
// however removing this will save you some instruction cycles and battery life (think so :) )
#define USING_MOD_LCD1x9

#if defined(USING_MOD_LCD1x9)
	#include "LCD1x9.h"
#else
	// empty define to prevent compile errors
	#define LCD1x9_WriteROM(x)
#endif


// enable this line to produce output hex to use with the "SerializatorDotNET.exe" program
#define OUTPUT_FOR_SERIALIZATOR_PROGRAM

/************************** VARIABLES ************************************/

#define SCAN_RETIRES 4
#define SEND_RETIRES 4

#define THERMOMETER   0x01

// This board can 
// upto 16 parameter types here!
typedef enum _PARAMETER_TYPE {
	PARAMETER_TYPE_UNKNOWN = 0,
	PARAMETER_TYPE_TEMPERATURE,
	PARAMETER_TYPE_DEVICENAME
} PARAMETER_TYPE;


/************************** VARIABLES ************************************/

/*************************************************************************/
// The variable myChannel defines the channel that the device
// is operate on. This variable will be only effective if energy scan
// (ENABLE_ED_SCAN) is not turned on. Once the energy scan is turned
// on, the operating channel will be one of the channels available with
// least amount of energy (or noise).
/*************************************************************************/
BYTE myChannel = 25;

// The name of the device (sent to server upon associating)
BYTE myDeviceName[16];

/*************************************************************************/
// AdditionalNodeID variable array defines the additional 
// information to identify a device on a PAN. This array
// will be transmitted when initiate the connection between 
// the two devices. This  variable array will be stored in 
// the Connection Entry structure of the partner device. The 
// size of this array is ADDITIONAL_NODE_ID_SIZE, defined in 
// ConfigApp.h.
// In this demo, this variable array is set to be empty.
/*************************************************************************/
#if ADDITIONAL_NODE_ID_SIZE > 0
    BYTE AdditionalNodeID[ADDITIONAL_NODE_ID_SIZE] = {THERMOMETER};
#endif

// states in the main FSM of the program
typedef enum _PROGRAM_STATES {
	PR_STATE_UNKNOWN = 0,
	PR_STATE_ERROR,
	PR_STATE_RESET,
	PR_STATE_INITIALIZE,
	PR_STATE_SEARCH_AND_CONNECT_RF,
	PR_STATE_SEND_DEVICENAME,
	PR_STATE_START_TEMPR,
	PR_STATE_READ_TEMPR,
	PR_STATE_START_BATTERY,
	PR_STATE_READ_BATTERY,
	PR_STATE_SEND_MSG_RF
} PROGRAM_STATES;

/* LOCAL FUNCTIONS DECLARATIONS */
static void InitializeAndTestThermometer(void);
static void FillInEUI(void); // fills in EUI address from somewhere (EEPROM for ex.)
static void FillInDeviceName(void); // fills in device name from somewhere (EEPROM for ex.)
static BYTE GetEEByte(UINT16 index); // get a byte from EEPROM

/*********************************************************************
* Function:         void main(void)
*
* Note:			    
**********************************************************************/
void main(void)
{   
	PROGRAM_STATES programState = PR_STATE_INITIALIZE;
	int sleepTime = 0; // in x4  seconds, if not required leave 0
	BYTE i, j;
	BYTE tmp;
    BYTE OperatingChannel = 0xFF;
    char retries;
    BOOL lowBatteryFlag = FALSE;
    BOOL enableLongSleep = FALSE; // flag to define that this should be the long sleep period (variability in period is applied as well as premature wake-up)
    WORD measuredTempr = 0;
    BOOL initializedFlag;
	
	/*******************************************************************/
	// Initialize the system
	/*******************************************************************/
	BoardInit();
	InitializeAndTestThermometer();

	#if defined(USING_MOD_LCD1x9)
		LCD1x9_Initialize();
	#endif
			
	// THE SUPER LOOP
	while(1) {
		enableLongSleep = FALSE;
		
		switch(programState) {
		case PR_STATE_INITIALIZE:
			initializedFlag = FALSE;
			// initialize connection from scratch
			FillInEUI();
			FillInDeviceName();
			MiApp_ProtocolInit(FALSE);
			
			retries	= SCAN_RETIRES;
			
			programState = PR_STATE_SEARCH_AND_CONNECT_RF;
			sleepTime = 0;
			break;
			
		case PR_STATE_SEARCH_AND_CONNECT_RF:
			LED_2 = 1;
		
			ConsolePutROMString((ROM char*)"\r\nStarting Active Scan...");
			LCD1x9_WriteROM("CONNECT..");
			
			// wake up the rf chip if necessary
			if(MiWiStateMachine.bits.Sleeping) {
				MiApp_TransceiverPowerState(POWER_STATE_WAKEUP);
			}	

			// search all channels at a duration 1s per channel
			j = MiApp_SearchConnection(10, 0xFFFFFFFF);
			
			if( j > 0 )	{
				// now print out the scan result.
				Printf("\r\nActive Scan Results: \r\n");
				for(i = 0; i < j; i++) {
					Printf("Channel: ");
					PrintDec(ActiveScanResults[i].Channel);
					Printf("   RSSI: ");
					PrintChar(ActiveScanResults[i].RSSIValue);
			#if defined(IEEE_802_15_4)
				#if ADDITIONAL_NODE_ID_SIZE > 0
					Printf("   PeerInfo: ");
					for(j = 0; j < ADDITIONAL_NODE_ID_SIZE; j++) {
						PrintChar(ActiveScanResults[i].PeerInfo[j]);
					}
				#endif
					Printf("    PANID: ");
					PrintChar(ActiveScanResults[i].PANID.v[1]);
					PrintChar(ActiveScanResults[i].PANID.v[0]);
					Printf("\r\n");
			#endif
					OperatingChannel = ActiveScanResults[i].Channel;
				}
			}
			
			LED_2 = 0;
			
			if( OperatingChannel != 0xFF ) {
				// select channel to operate on
				MiApp_SetChannel(OperatingChannel);

	        	// display info about the connection on the console
				DumpConnection(0xFF);
	  
	  			MiApp_ConnectionMode(ENABLE_ALL_CONN);
				i = MiApp_EstablishConnection(0, CONN_MODE_DIRECT);
				Printf("Connected on Channel ");
				PrintDec(myChannel);
				Printf("\r\n");
				LCD1x9_WriteROM("CONNECTED");
	
				retries	= SEND_RETIRES;
				initializedFlag = TRUE;
	
				programState = PR_STATE_SEND_DEVICENAME;
				sleepTime = 0;
			} else {
				if(--retries < 0) {
					Printf("\r\nERROR: CANNOT CONNECT TO ANY NETWORK!!!");
					programState = PR_STATE_RESET;
					enableLongSleep = TRUE;
					sleepTime = 800;
					
				} else {
					Printf("\r\nNo Suitable PAN, Rescanning...");
					programState = PR_STATE_START_TEMPR;
					sleepTime = 0;
				}
			}	
			break;
			
		case PR_STATE_SEND_DEVICENAME:
			LED_2 = 1;
			
			MiApp_FlushTx();
			tmp = (PARAMETER_TYPE_DEVICENAME & 0x0F);
			if(lowBatteryFlag) tmp |= 0x10;
			MiApp_WriteData(tmp);
			for(i = 0; i < 16; i++)	
				MiApp_WriteData(myDeviceName[i]);

			// send msg to parent using security
			if( MiApp_UnicastConnection(0, TRUE) ) {
				Printf("\r\nName sent!\r\n");
				
				programState = PR_STATE_START_TEMPR;
				sleepTime = 0;
			} else {
				Printf("\r\nERROR: Name sending failed!\r\n");
				if(--retries < 0) {
					programState = PR_STATE_INITIALIZE;
					sleepTime = 0;					
				} else {
					programState = PR_STATE_SEND_DEVICENAME;
					enableLongSleep = TRUE;
					sleepTime = 20;
				}
			}	
						
			LED_2 = 0;
			break;
			
		case PR_STATE_START_TEMPR:
			if(TCN75A_StartSample() == -1) {
				Printf("\r\nERROR: TEMPERATURE SAMPLING!!!");
				programState = PR_STATE_ERROR;
				sleepTime = 0;
			} else {
				programState = PR_STATE_START_BATTERY;
				sleepTime = 1;
			}	
			break;
			
		case PR_STATE_READ_TEMPR:
			measuredTempr = TCN75A_ReadTempr();
			
			if(initializedFlag)
				programState = PR_STATE_SEND_MSG_RF;
			else
				programState = PR_STATE_READ_BATTERY; // if not connected at least read the battery
			
			sleepTime = 0;
			break;
				
		case PR_STATE_SEND_MSG_RF:
			LED_2 = 1;
			
			LCD1x9_WriteROM("SENDING");
			
			// wake up the rf chip if necessary
			if(MiWiStateMachine.bits.Sleeping) {
				MiApp_TransceiverPowerState(POWER_STATE_WAKEUP);
			}	
			
			// fill in data to send
			// format goes like this:
			// |xxxxABBBB|CCCCCCCC|CCCCCCCC|
			// A - 1 if battery is low
			// B - parameter type (only temperature for the time being!)
			// C - parameter value
			MiApp_FlushTx();
			tmp = (PARAMETER_TYPE_TEMPERATURE & 0x0F);
			if(lowBatteryFlag) tmp |= 0x10;
			MiApp_WriteData(tmp);
			MiApp_WriteData( *((BYTE*)&measuredTempr + 1) );
			MiApp_WriteData( *((BYTE*)&measuredTempr) );

			// send msg to parent using security
			if( MiApp_UnicastConnection(0, TRUE) ) {
				Printf("\r\nMsg sent!\r\n");
				
				programState = PR_STATE_READ_BATTERY;
				sleepTime = 0;
			} else {
				Printf("\r\nERROR: Msg sending failed!\r\n");
				if(--retries < 0) {
					programState = PR_STATE_INITIALIZE;
					sleepTime = 0;					
				} else {
					programState = PR_STATE_READ_BATTERY;
					sleepTime = 0;
				}
			}	
									
			LED_2 = 0;
			break;

		case PR_STATE_START_BATTERY:
			PIR2bits.HLVDIF = 0;
			HLVDCONbits.HLVDEN = 1; // enable module
					
			// do not wait, time till reading should be pretty much enough!
			//while(!HLVDCONbits.IRVST); // wait for stable operation
						
			programState = PR_STATE_READ_TEMPR;
			sleepTime = 0;
			break;
			
		case PR_STATE_READ_BATTERY:
			HLVDCONbits.HLVDEN = 0; // disable module
			
			lowBatteryFlag = PIR2bits.HLVDIF;
			
			if(initializedFlag) {
				programState = PR_STATE_START_TEMPR;
				enableLongSleep = TRUE;
				sleepTime = 30;
			} else {
				programState = PR_STATE_SEARCH_AND_CONNECT_RF;
				enableLongSleep = TRUE;
				sleepTime = 120;
			}			
			break;

		case PR_STATE_ERROR:
			for(i = 0; i < 5; i++) {
				LED_1 = 1;
				DelayMs(500);
				LED_1 = 0;
				DelayMs(500);
			}
			
			// put off everything
			MiApp_TransceiverPowerState(POWER_STATE_SLEEP);
			LED_1 = 0;
			LED_2 = 0;
			
			Sleep();
			Reset();
			break;
		case PR_STATE_RESET:
			Reset();

		// should never get here
		default:
		case PR_STATE_UNKNOWN:
			Reset();
		}
		
		// do smth about received messages
		if(!MiWiStateMachine.bits.Sleeping) {
			if( MiApp_MessageAvailable() )
				MiApp_DiscardMessage();
		} 
		
		// sleep some time
		if(sleepTime > 0) {
			LED_2 = 0;
			
			// put rf chip to sleep, if not already sleeping
			if(!MiWiStateMachine.bits.Sleeping)
				MiApp_TransceiverPowerState(POWER_STATE_SLEEP);
			
			// update temperature and battery status on display
			// ex: "24.5oC LB"
			#if defined(USING_MOD_LCD1x9)
			{
				BYTE tmp;
				char str[10];
				
				tmp = (BYTE)(measuredTempr >> 8);
				str[0] = (tmp / 10) | 0x30;
				str[1] = (tmp % 10) | 0x30;
				str[2] = '.';
				tmp = (BYTE)((measuredTempr >> 3) & 0x01F);
				if(tmp > 29)
					str[3] = '9';
				else
					str[3] = (tmp / 3) | 0x30;
				
				str[4] = '*';
				str[5] = 'C';
				str[6] = ' ';
				
				if(lowBatteryFlag) {
					str[7] = 'L';
					str[8] = 'B';
					str[9] = 0;
				} else
					str[6] = 0;
					
				LCD1x9_Write(str);
			}
			#endif
			
			if(enableLongSleep) {
				// randomize sleep period wih +/- 15s max
				// use timer value and temperature value as a source of entropy
				tmp = TMR0L + ((BYTE)measuredTempr >> 4);
				if(tmp & 0x10)
					tmp |= 0xF0;
				else
					tmp &= 0x0F;
				
				sleepTime += (char)tmp;
				if(sleepTime < 0)
					sleepTime = 1;
				
				// enable interrupts on change to wake up device from sleep
				tmp = PORTB; // to be able to clear interrupt flag, first port has to be read
				INTCONbits.RBIF = 0;
				INTCONbits.RBIE = 1;
			}	
	
			// run at 32kHz to save even more battery
			OSCCON = 0b10000010;
			WDTCONbits.SWDTEN = 1;
			
			while(sleepTime--) {
				// enable watch dog timer
				ClrWdt();
				
				// make a blink every 2nd second
				if(lowBatteryFlag && (sleepTime & 0x01)) {			
					LED_1 = 1; // display battery state from prevous state

					tmp = 0x10;
					while(--tmp)
						Nop();					
					LED_1 = 0; // turn off indicator for low battery
				}	
				
				Sleep();
				
				// if wakup caused by IOC then don't sleep any more cycles
				if(INTCONbits.RBIF) {
					sleepTime = 0;
					LED_2 = 1; // indicate the request is accepted					
				}
			}
			
			// disable watch dog timer
			WDTCONbits.SWDTEN = 0;
			OSCCON = 0b01110010; // again @16MHz
		}	
	}
}

static void InitializeAndTestThermometer(void)
{
	if( TCN75A_Initialize() != 0) {
		char i;
		// blink 3 times to indicate error in thermometer
		for(i = 0; i < 3; i++) {
			LED_1 = 1;
			DelayMs(500);
			LED_1 = 0;
			DelayMs(500);
		}

		Sleep();
		Reset();
//		while(1);
	}	
}

/* UNIQUE DEVICE ADDRESS AND NAME */
// if using the serializer program these valuses have to excluded from output hex file
#ifndef OUTPUT_FOR_SERIALIZATOR_PROGRAM
#pragma romdata eedata=0xF00000
	rom far unsigned char myLongAddressEE[8] = {EUI_0,EUI_1,EUI_2,EUI_3, EUI_4, EUI_5,EUI_6,EUI_7};
	rom far unsigned char myDeviceNameEE[17] = ZIGBEE_DEVICE_NAME;
#pragma code
#else
	extern rom far unsigned char myLongAddressEE[8];
	extern rom far unsigned char myDeviceNameEE[17];
#endif

static BYTE GetEEByte(UINT16 index)
{
	EEADR = (BYTE)index;
	EEADRH = 0;//*((BYTE*)(&index + 1));
	
	EECON1bits.EEPGD = 0;
	EECON1bits.CFGS = 0;
	EECON1bits.RD = 1;
	
	return EEDATA;
}

static void FillInEUI(void)
{
	extern BYTE myLongAddress[];
	
	char i;
	
	for(i = 0; i < 8; i++)	
		myLongAddress[i] = GetEEByte(0 + i);
}

static void FillInDeviceName(void)
{
	char i;
	
	for(i = 0; i < 16; i++)	
		myDeviceName[i] = GetEEByte(sizeof(myLongAddressEE) + i);	
}


	
